from pmlab_lite.log import EventLog
from pmlab_lite.manipulable import *


class EventLogFilters(Filter):
    def __init__(self, log: EventLog):
        self.log = log


class FilterPerAttribute(EventLogFilters):
    def __call__(self, name: str, value) -> EventLog:
        f = EventLog()
        for e in self.log:
            if isinstance(value, list):
                if e[name] in value:
                    f.add_event(e)
            else:
                if e[name] is value:
                    f.add_event(e)
        self.log = f
        return self.log


class ProjectActivities(EventLogFilters):
    def __call__(self, activities_to_keep) -> EventLog:
        f = EventLog()
        for e in self.log:
            if e.get_activity_name() in activities_to_keep:
                f.add_event(e)
        self.log = f
        return self.log


class TracesStartingWith(EventLogFilters):
    def __call__(self, starting_activity) -> EventLog:
        f = EventLog()
        for t in self.log.get_traces():
            if t[0].get_activity_name() is starting_activity :
                for e in t:
                    f.add_event(e)
        self.log = f
        return self.log


class TracesEndingWith(EventLogFilters):
    def __call__(self, ending_activity) -> EventLog:
        f = EventLog()
        for t in self.log.get_traces():
            if t[-1].get_activity_name() is ending_activity :
                for e in t:
                    f.add_event(e)
        self.log = f
        return self.log


class TracesWithDirectFollowing(EventLogFilters):
    def __call__(self, first_activity, second_activity) -> EventLog:
        f = EventLog()
        for t in self.log.get_traces():
            contains_df = False
            for e1, e2 in zip(t[:-1], t[1:]):
                if e1.get_activity_name() is first_activity and e2.get_activity_name() is second_activity:
                    contains_df = True
                    break
            if contains_df:
                for e in t:
                    f.add_event(e)
        self.log = f
        return self.log



from dateutil.parser import parse
import gzip
import xml.etree.ElementTree as xmltree
from pmlab_lite.log import *


def import_from_xes(file):
    """Load a log from the XES format.

    [filename] can be a file or a filename."""
    if isinstance(file, str):
        filename = file
        if filename.endswith('.gz'):
            file = gzip.open(filename, 'rb')
        else:
            pass  # Just send the filename to xmltree.parse
    else:
        filename = file.name

    # construct the new event log  object
    imported_log = EventLog()

    # identify all traces in the log (assuming just one process per log)
    ns = {'xes': 'http://www.xes-standard.org/'}
    tree = xmltree.parse(file)
    root = tree.getroot()
    traces = root.findall('xes:trace', ns)

    # parse all traces in the log
    for t in traces:
        # extract the case id
        case_id = t.find('xes:string[@key="concept:name"]', ns).attrib['value']
        # parse all events in the trace
        for e in t.findall('xes:event', ns):
            # extract the activity name and construct the new event
            activity_name = e.find('xes:string[@key="concept:name"]', ns).attrib['value']
            event = Event(activity_name, case_id)
            # parse all string attributes
            for a in e.findall('xes:string', ns):
                # we need to exclude the name of the event
                if a.attrib['key'] != "concept:name":
                    event[a.attrib['key']] = a.attrib['value']
            # parse all integer attributes
            for a in e.findall('xes:int', ns):
                event[a.attrib['key']] = int(a.attrib['value'])
            # parse all date attributes
            for a in e.findall('xes:date', ns):
                event[a.attrib['key']] = parse(a.attrib['value'])
            # parse all boolean attributes
            for a in e.findall('xes:boolean', ns):
                event[a.attrib['key']] = a.attrib['value'].lower() == 'true'
            imported_log.add_event(event)
    return imported_log


def print_log(l):
    for t in l.get_traces():
        print("TRACE:")
        for e in t:
            print("> ", e.get_activity_name(), e.get_case_id(), e.values())



# import sys
from abc import ABC, abstractmethod


# import pmlab_lite


class AbstractPetriNet(ABC):
	"""This abstract class defines a blueprint for petri net
	implementations. All concrete classes should implement the
	given functions.
	"""

	def __init__(self):
		self.places = {}
		self.transitions = {}
		self.edges = []
		self.marking = []
		self.capacity = []
		self.counter = 0  # mapping

	@abstractmethod
	def add_place(self, name, capacity=1):
		pass

	@abstractmethod
	def remove_place(self, name):
		pass

	@abstractmethod
	def add_transition(self, name):
		pass

	@abstractmethod
	def remove_transition(self, name):
		pass

	@abstractmethod
	def add_edge(self, place, transition, two_way=False):
		pass

	@abstractmethod
	def remove_edge(self, place, transitions):
		pass

	@abstractmethod
	def is_enabled(self, transition):
		pass

	@abstractmethod
	def add_marking(self, place, token=1):
		pass

	@abstractmethod
	def replay(self):
		pass



from pmlab_lite.pn import PetriNet
from .alignment import Alignment
from . import variables as v
import numpy
import heapq

BLANK = '>>' 
EPSILON = 0.00001


def store_variables(incidence_matrix, init_mark_vector, final_mark_vector, transitions_by_index):
	v.incidence_matrix = incidence_matrix
	v.init_mark_vector = init_mark_vector
	v.final_mark_vector = final_mark_vector
	v.transitions_by_index = transitions_by_index
	v.solutions = []


class A_star(Alignment):
	
	def __init__(self):
		Alignment.__init__(self)
		self.solutions = []
		
	#a star algorithm
	def A_star_exe(self, synchronous_product, trace, heuristic_variant="lp", no_of_solutions=1):
		v.open_list = open_list = []
		v.closed_list = closed_list = []

		heuristic = Heuristic(heuristic_variant)
		incidence_matrix = synchronous_product.incidence_matrix()
		transitions_by_index = synchronous_product.transitions_by_index()
		init_mark_vector = synchronous_product.get_spnets_initial_marking() 
		final_mark_vector = synchronous_product.get_spnets_final_marking()
			
		store_variables(incidence_matrix, init_mark_vector, final_mark_vector, transitions_by_index)
			
		#creating the initial node in the search space
		current_node = Node()
		current_node.Make_initial_node(trace, heuristic)
		open_list.append( [current_node.total_cost, current_node] )
		
		#number for drawing the search space in order
		count = 0
		#iterating until open_list has no elements
		while len(open_list) > 0:
			# build min heap from open list
			heapq.heapify(open_list)					

			current_node = heapq.heappop(v.open_list)	#pop cheapest node from open list as (total cost, node)
			closed_list.append(current_node)		#append it to closed list as (total cost, node)
			current_node = current_node[1]			#make the current node just the node of the tuple
			current_node.number = count
			
			#checking whether the node is a solution. else: investigate
			if( numpy.array_equal( current_node.marking_vector, final_mark_vector) ):
				self.solutions.append(current_node)
				v.solutions.append(self.alignment_move)
					
				if len(self.solutions) >= no_of_solutions:
					break
			else:
				current_node.Investigate(incidence_matrix, transitions_by_index, heuristic)
				count += 1
		
		self.closed_list_end = closed_list
		self.__Fitness()
		self.__Move_alignment()
		self.__Move_in_model()
		self.__Move_in_log()
		return v.solutions
	
	def __Fitness(self):
		for sol in self.solutions:
			u = sol.alignment_up_to
			self.fitness.append( round ( len( [e for e in u if ((e[0]!=BLANK and e[1]!=BLANK) or ('tau' in e[0])) ]) / len(u), 3) )
		
	def __Move_alignment(self):
		for sol in self.solutions:
			u = sol.alignment_up_to
			self.alignment_move.append( [e for e in u if ('tau' not in e[0])] )
		
	def __Move_in_model(self):
		for sol in self.solutions:
			u = sol.alignment_up_to
			self.move_in_model.append( [e[0] for e in u if (e[0]!=BLANK and ('tau' not in e[0]))] )  
		
	def __Move_in_log(self):
		for sol in self.solutions:
			u = sol.alignment_up_to
			self.move_in_log.append( [e[1] for e in u if e[1] != BLANK])								


class Node():	
	def __init__(self):
		#shows a marking node
		self.number = 0
		self.marking_vector = []
		self.active_transitions = []
		self.parent_node = ''
		self.observed_trace_remain = []
		self.alignment_up_to = []			#it's of the form [ (t1,t1), (t2,-), (-,t3), ... ]
		
		self.cost_from_init_marking = 1 * sum( [EPSILON if ((x[0]!=BLANK and x[1]!=BLANK) or ('tau' in x[0])) else 1 for x in self.alignment_up_to] )
		self.cost_to_final_marking = 1000
		self.total_cost = self.cost_from_init_marking + self.cost_to_final_marking
	
	#method for comparing two objects (python 3.7)
	def __lt__(self, other):
		return self.total_cost < other.total_cost
		
	#this funtion calls other functions to investigate the current node
	def Investigate(self, incidence_matrix, transitions_by_index, heuristic):
		'''
		:param incidence_matrix: matrix of the synchronous product
		:param transitions by index: reverse of the dict 'net.transitions', to access transitions names by index
		'''

		self.Find_active_transitions(incidence_matrix)

		#heuristic evaluation of active transitions
		for i in self.active_transitions:
			#make child node and update it's marking, i.e. the current marking after transition i was fired
			child_node = Node()
			child_node.marking_vector = incidence_matrix[:, i] + self.marking_vector
			child_node.parent_node = self

			# --Synchronous move--
			if transitions_by_index[i].endswith("synchronous"):
				
				#update it's remaining trace
				child_node.observed_trace_remain = self.observed_trace_remain[1:]
				child_node.alignment_up_to = self.alignment_up_to + [ (transitions_by_index[i][:-12], transitions_by_index[i][:-12] ) ]
				
			# --Model       move--
			elif transitions_by_index[i].endswith("model"):
				
				#update it's remaining trace
				child_node.observed_trace_remain = self.observed_trace_remain[:]
				child_node.alignment_up_to = self.alignment_up_to + [ ( transitions_by_index[i][:-6], BLANK ) ]
					
			# --Log         move--
			elif transitions_by_index[i].endswith("log"):

				#update it's remaining trace
				child_node.observed_trace_remain = self.observed_trace_remain[1:]
				child_node.alignment_up_to = self.alignment_up_to + [ (BLANK, transitions_by_index[i][:-4] ) ]

			#update the child nodes costs
			child_node.Update_costs(heuristic)
	
			#check whether it's in the closed list or it's a cheaper version of same marking
			child_node.Add_node(v.open_list, v.closed_list)
			
	def Find_active_transitions(self, incidence_matrix):
		#looping over transitions of the synchronous product, to see which are active, given the marking of that node
		for i in range(0, incidence_matrix.shape[1]):											#im.shape[1] returns #columns = #transitions
			
			if numpy.all( (incidence_matrix[:, i] + self.marking_vector) >= 0 ):				#im[:,i] returns i-th column as list
				#transition i is active
				if i not in self.active_transitions:
					self.active_transitions.append(i)
					
	#deciding on whether or not to add a node to the open list
	def Add_node(self, open_list, closed_list, id):
		#checking whether it is in the closed list
		#ind is a list like [12,34,10]
		ind = [k for k in range( len(closed_list) ) if numpy.array_equal(self.marking_vector, closed_list[k][1].marking_vector) ]
		
		if len(ind) > 0:
			pass
		#checking whether it is in the open list, update if we found it
		else:
			ind = [k for k in range(len(open_list)) if numpy.array_equal(self.marking_vector, open_list[k][1].marking_vector)]
			
			#at least once in open list
			if ind:
				for k in ind:
					if open_list[k][1].cost_from_init_marking > self.cost_from_init_marking:	
						open_list[k] = [self.total_cost, self]
					else: 
						continue
			#not in open list yet
			else:
				open_list.append( [self.total_cost, self] )

	def Cost_from_init(self):
		#base_cost = 1 + EPSILON
		#add base cost into initial cost calculation
		self.cost_from_init_marking = 1 * sum( [EPSILON if ((x[0]!=BLANK and x[1]!=BLANK) or ('tau' in x[0])) else 1 for x in self.alignment_up_to] )

	#calcucalte the remaining cost to the final marking,based on a heuristic
	def Cost_to_final(self, heuristic):
		heuristic.heuristic_to_final(self)

	def Update_costs(self, heuristic):
		self.Cost_from_init()
		self.Cost_to_final(heuristic)	
		self.total_cost = self.cost_from_init_marking + self.cost_to_final_marking
		
	def Make_initial_node(self, trace, heuristic):
		self.marking_vector = v.init_mark_vector
		self.observed_trace_remain = trace    
		self.Cost_to_final(heuristic)
		return self


class Heuristic():

	def __init__(self, variant):
		self.variant = variant

	def heuristic_to_final(self, node):
		if self.variant == "tl":
			self.remaining_trace_length_heursitic(node)
		elif self.variant == "lp":
			self.linear_programming_heursitic(node)
	
	def remaining_trace_length_heursitic(self, node):
		node.cost_to_final_marking = len(node.observed_trace_remain) * EPSILON

	def linear_programming_heursitic(self, node):
		#Heuristic.heurisitic_to_final()
		b = numpy.array(v.final_mark_vector) - numpy.array(node.marking_vector) 
		x = numpy.linalg.lstsq(v.incidence_matrix, b, rcond=None)[0]
		
		# important note: with rcond=None, the the results in costs calculated my vary on different machines
		# on three machines tested the unrounded sum of each x was mostly equal, differences occured in +-10 digits after comma
		# after rounding up or down though, differences where bigger on each machine (e.g. machine1 rounds -0.000001 to 0
		# 			and machine2 rounds 0.00000001 to 1 -> making the sums differ in multiples of 1
		#with rcond=-1 the same results have been calculated on each machine
		
		#rounding up or down, 
		x[x > 0] = 1
		x[x <= 0] = 0
		
		for key in v.transitions_by_index:
			if v.transitions_by_index[key].startswith('tau'):
				x[key] = 0
		
		node.cost_to_final_marking = numpy.sum(x)



from functools import reduce
from pmlab_lite.discovery import ProcessTree, Cut
from pmlab_lite.log import EventLog
from pmlab_lite.pn import PetriNet


class InductiveMiner():

	def __init__(self, log: EventLog):
		self.log = self.prepare_log(log)

	def prepare_log(self, event_log: EventLog) -> list:
		"""
		Convert EventLog into simple list structure.
		:param event_log: EventLog
		:return: list
		"""

		log = []
		for trace in event_log.get_traces():
			red_trace = []
			[red_trace.append(event['activity_name']) for event in trace]
			log.append(red_trace)
		return log

	def discover(self) -> ProcessTree:
		return ProcessTree(self, self.log)

	def split_log(self, split: list, cut, log: list):
		log_left = []
		log_right = []

		if cut == Cut.SEQ:
			# max sequence cut
			logs = []
			for part in split:
				sub_log = []

				if len(part) > 1 and \
						all([isinstance(sub, list) for sub in part]):
					part = list(reduce(lambda x, y: x + y, part))

				for trace in log:
					if any([ele in trace for ele in part]):
						part_of_trace = list(filter(lambda x: x in part, trace))
						if part_of_trace not in sub_log:
							sub_log.append(part_of_trace)
					else:
						if [] not in sub_log:

							sub_log.append([])
				logs.append(sub_log)

			return logs

		elif cut == Cut.PARA:
			for trace in log:
				if split[0][0] in trace and split[1][0] in trace:
					left_part = [split[0][0]]
					right_part = [split[1][0]]
				elif split[0][0] in trace:
					left_part = [split[0][0]]
					right_part = []
				else:
					left_part = []
					right_part = [split[1][0]]

				if right_part not in log_right:
					log_right.append(right_part)

				if left_part not in log_left:
					log_left.append(left_part)

		elif cut == Cut.EXCL:
			for trace in log:
				if any([ele in trace for ele in split[0]]):
					log_left.append(trace)
				else:
					log_right.append(trace)

		elif cut == Cut.LOOP:
			for trace in log:
				if any([ele in trace for ele in split[1]]):
					left_part = split[0]
					right_part = split[1]
				else:
					left_part = trace
					right_part = []

				if right_part not in log_right:
					log_right.append(right_part)

				if left_part not in log_left:
					log_left.append(left_part)

		return log_left, log_right

	def tree_to_petri_net(self, tree: ProcessTree) -> PetriNet:
		net = PetriNet()
		net.add_place(1)
		# net.add_marking(1, 1)

		def add_region(sub_tree, place_in, place_out):

			if isinstance(sub_tree, ProcessTree):
				if sub_tree.cut == Cut.PARA:
					# add additional input place
					add_region(sub_tree.children[0], place_in, place_out)
					for child in sub_tree.children[1:]:
						new_place_in = max(list(net.places.values()))[0] + 1
						add_region(sub_tree.children[0], new_place_in,
								   place_out)

				elif sub_tree.cut == Cut.EXCL:
					pass
				elif sub_tree.cut == Cut.SEQ:
					pass
				elif sub_tree.cut == Cut.LOOP:
					pass
				else:
					next = min(list(net.transitions.values()))[0] - 1
					net.add_transition(''.join(['sub region', str(next)]))
					trans = min(list(net.transitions.values()))[0]
					net.add_edge(place_in, trans)
					net.add_edge(trans, place_out)
			else:
				net.add_transition(sub_tree)
				trans = min(list(net.transitions.values()))[0]
				net.add_edge(place_in, trans)
				net.add_edge(trans, place_out)



		for child in tree.children:
			pre_place = max(list(net.places.values()))
			net.add_place(pre_place + 1)
			add_region(child, pre_place, pre_place + 1)



		return net


import sys
from pmlab_lite.log import *


class Manipulable:
    """ This class represents an object that can be manipulated. """

    def __getattr__(self, attname):
        class_ = getattr(sys.modules["pmlab_lite.log.transformations"], attname)
        return class_(self)


class Filter:
    """ This class represents a filter that can be used to manipulate a manipulable. """

    def __call__(self, caller):
        raise NotImplemented



from pmlab_lite.manipulable import Manipulable


class Event(dict):
    """"
    An event is the minimum observable unit of information. It actually is a dictionary with, at least, two attributes:
    'activity_name' and 'case_id'.
    """
    def __init__(self, activity_name, case_id):
        self["activity_name"] = activity_name
        self["case_id"] = case_id

    def get_activity_name(self):
        return self["activity_name"]

    def get_case_id(self):
        return self["case_id"]

    def get_attributes(self):
        return self.keys()


class EventCollection(Manipulable):
    """"
    An event collection is an abstract iterable collection of events. This is the smallest collection of events
    """
    def __iter__(self):
        return self

    def __next__(self) -> Event:
        raise NotImplemented


class EventLog(EventCollection):
    """"
    An event log is a collection of events with a fixed size. It can also extend the 'Manipulable' (or whatever we call
    it) class. With this implementation, it is not possible to have trace attributes, but only events attributes.
    Direct access to single traces is also possible
    """
    def __init__(self):
        self.events = list() # events are still stored as list of events
        self.traces = dict() # traces are stored as dictionary of case id -> list of events

    def add_event(self, event: Event):
        self.events.append(event)
        if event.get_case_id() in self.traces:
            self.traces[event.get_case_id()].append(event)
        else:
            self.traces[event.get_case_id()] = [event]
        return self

    def add_trace(self, case_id, activity_names: []):
        for activity_name in activity_names:
            self.add_event(Event(activity_name, case_id))
        return self

    def get_traces(self):
        return self.traces.values()

    def get_trace(self, case_id):
        return self.traces[case_id]

    def __iter__(self):
        return self.events.__iter__()

    def __next__(self):
        return self.events.__iter__().__next__()


import graphviz as gr
from graphviz import Digraph
from pmlab_lite.discovery.cut import Cut
from pmlab_lite.discovery.process_tree import ProcessTree
from pmlab_lite.helper.graph import Graph
from pmlab_lite.pn import AbstractPetriNet
from pmlab_lite.pn import PetriNet

BLANK = '>>'

def get_marking(marking: int):
	"""viz import
	Create label for a place with its corresponding number of tokens.

	:param marking: number of token
	:return: label
	"""

	if marking < 4:
		token = '&#11044;'
		if marking == 0:
			return ''
		elif marking == 1:
			return '<<B>%s</B>>' % token
		elif marking == 2:
			return '<<B>%s %s</B>>' % (token, token)
		else:
			return '<<B>%s<BR/>%s %s</B>>' % (token, token, token)
	else:
		return '4'

def draw_petri_net(input_net: AbstractPetriNet, filename="petri_net",
				   format="pdf"):
	"""
	This function transforms the given Petri net structure
	into DOT notation and returns it as a digraph object. Call
	the render() function to save it to file.

	Args:
		filename: string name of the file
		format: export format

	Retruns:
		Digraph object
	"""
	
	dot = Digraph(name=filename, format=format)
	dot.attr(rankdir='LR', fontsize="10", nodesep="0.35",
			 ranksep="0.25 equally")

	# draw transitions
	dot.attr('node', shape='box', penwidth="1", fontsize="10",
			 fontname="Helvetica")

	for key, values in input_net.transitions.items():
		if key != 'tau' and not key.endswith("_synchronous"):
			for t in values:
				dot.node(str(t), key)
					
	# draw sync transitions
	dot.attr('node', shape='box', penwidth="1", fontsize="10",
			 fontname="Helvetica", color='green')
			 
	for key, values in input_net.transitions.items():
		if key.endswith("_synchronous"):
			for t in values:
				dot.node(str(t), key)

	if 'tau' in input_net.transitions.keys():
		# draw tau transistions
		dot.attr('node', shape='square', style="filled",
				 color='black', penwidth="1", fontsize="5",
				 fontname="Helvetica")

		for t in input_net.transitions['tau']:
			dot.node(str(t))

	# draw place
	dot.attr('node', shape='circle', penwidth='1', fontsize='10',
			 fontname='Helvetica', style='unfilled', color='black')

	# draw marking
	for id, p in input_net.places.items():
		label = get_marking(input_net.marking[id])
		dot.node(str(p), label=label, xlabel=str(p))

	# draw edges
	#TODO get colored edges for sync prod edges
	for e in input_net.edges:
		dot.edge(str(e[0]), str(e[1]))

	render_dot(dot, filename)  # TODO: change to just return the object
	return dot

def draw_synchronous_product(input_net: AbstractPetriNet, filename="synchronous_product_net", format="pdf"):
	color = ["black", "chartreuse3", "yellow1"]
	transitions_by_index = input_net.transitions_by_index()

	dot = gr.Digraph(name=filename, format=format)
	dot.attr(rankdir='LR', fontsize="10", nodesep="0.35", ranksep="0.25 equally" )

	#draw places
	for id, p in input_net.places.items():
		label = get_marking(input_net.marking[id])
		dot.node(str(p), label=label, xlabel=str(p), shape="circle")

	#draw transitions	
	for i in range(0,len(transitions_by_index)):
		if transitions_by_index[i].endswith("_model"):
			dot.node( str(-(i+1)), "(" + transitions_by_index[i][:-6] + "," + BLANK + ")",shape="rect", style='unfilled', color=color[0])
		elif transitions_by_index[i].endswith("_log"):
			dot.node( str(-(i+1)), "(" + BLANK + "," + transitions_by_index[i][:-4] + ")", shape="rect", style='filled')
		elif transitions_by_index[i].endswith("_synchronous"):
			dot.node( str(-(i+1)), "(" + transitions_by_index[i][:-12] + "," + transitions_by_index[i][:-12] + ")", shape="rect", style='filled', color=color[1])


	#draw edges
	for e in input_net.edges:
	#the edge is coming from a transition
		if e[0] < 0:
			if transitions_by_index[-(e[0]+1)].endswith("_synchronous"):
				dot.edge( str(e[0]), str(e[1]), color=color[1] )
			else:
				dot.edge( str(e[0]), str(e[1]) )
	#the edge is going to a transition
		else:
			if transitions_by_index[-(e[1]+1)].endswith("_synchronous"):
				 dot.edge( str(e[0]), str(e[1]), color=color[1] )
			else:
				dot.edge( str(e[0]), str(e[1]) )

	#f = open("./sync_product.dot", "w")
	#f.write(dot.source)

	render_dot(dot, filename)  # TODO: change to just return the object
	return dot

def draw_a_star_search_space(alignment, filename="search_space", format="pdf"):
    closed_list = alignment.closed_list_end
    initial_mark_vector = closed_list[0][1].marking_vector
    final_mark_vector = alignment.solutions[0].marking_vector
    colors = ["indianred2", "darkseagreen", "gray"]
    counter = 100

    dot = Digraph(name=filename, format=format)
    dot.attr(rankdir='LR', fontsize="3", nodesep="0.35",
			 ranksep="0.25 equally")
    
    #draw nodes
    dot.attr('node', shape='circle', penwidth='1', fontsize='12',
			 fontname='Helvetica', style='filled')
    fontcolor = 'black'
    changed_fontcolor = False
    for e in closed_list:
        #change color for darker nodes
        if not changed_fontcolor and counter < 30:
            changed_fontcolor = True
            fontcolor = 'white'
        #node is the final marking
        if ( np.array_equal(e[1].marking_vector, final_mark_vector) ):
            dot.node(str(e[1].marking_vector), label="H = " + str(round(float(e[1].cost_to_final_marking),3)) + "\nG = " + str(round(float(e[1].cost_from_init_marking),3)), 
                color=colors[1])
        #node is the initial marking
        elif ( np.array_equal(e[1].marking_vector, initial_mark_vector) ):
            dot.node(str(e[1].marking_vector), label="H = " + str(round(float(e[1].cost_to_final_marking),3)) + "\nG = " + str(round(float(e[1].cost_from_init_marking),3)),
                color=colors[0])
        #any other node
        else:
            dot.node(str(e[1].marking_vector), label="H = " + str(round(float(e[1].cost_to_final_marking),3)) + "\nG = " + str(round(float(e[1].cost_from_init_marking),3)), 
                color=colors[2]+str(counter), fontcolor=fontcolor)
        if counter > 0:
            counter -= 1

    #draw edges
    for node in closed_list:
        #very first nodes parent node is " '' "
        if node[1].parent_node != '':
            parent_node = node[1].parent_node
            dot.edge(str(parent_node.marking_vector), str(node[1].marking_vector), label=str(node[1].number))

    render_dot(dot,filename)
    return dot


def draw_graph(graph: Graph, filename="graph", format="pdf", render=False,
			   view=True):
	dot = Digraph(name=filename, format=format)

	dot.attr(rankdir='LR', fontsize="10", nodesep="0.35",
			 ranksep="0.25 equally")

	# draw nodes
	dot.attr('node', penwidth="1", fontsize="10",
			 fontname="Helvetica")

	for node in graph.vertexes:
		dot.node(str(node))

	# draw edges
	for node in graph.vertexes:
		for target in graph.vertexes[node]:
			dot.edge(str(node), str(target))

	if render:
		render_dot(dot, filename, view)

	return dot

def draw_process_tree(tree: ProcessTree, name='process_tree', format='png',
					  render=True):
	"""
	This function transforms the given process tree
	into DOT notation renders it and returns it as a digraph object.

	:param tree: process tree object
	:param name: file name
	:param format: file format

	:return: Digraph object
	"""

	def gen_label(cut: int):
		"""
		Return node symbol, depending on cut type
		:param cut: cut code

		:return str: symbol
		"""

		if cut == Cut.SEQ:
			return '&#10140;'
		elif cut == Cut.PARA:
			return '&#43;'
		elif cut == Cut.EXCL:
			return '&#215;'
		elif cut == Cut.LOOP:
			return '&#x21BB;'

	def draw_children(parent: str, child, dot: Digraph):

		def draw_node(sub_tree, dot: Digraph):
			dot.attr('node', shape='circle', penwidth="1", fontsize="10",
					 fontname="Helvetica", height="0.3", fixedsize="false")

			node_name = ''.join([str(sub_tree.parent), str(sub_tree.id)])
			dot.node(node_name, label=gen_label(sub_tree.cut))

			for child in sub_tree.children:
				draw_children(node_name, child, dot)

		def draw_leaf(parent, leaf, dot: Digraph):
			dot.attr('node', shape='box', color='black', penwidth="1",
					 fontsize="10",
					 fontname="Helvetica", height="0.5", fixedsize="false")

			leaf_name = ''.join([parent, str(leaf)])
			if leaf == 'tau':
				dot.node(leaf_name, label=str(leaf), shape="square",
						 style='filled')
			else:
				dot.node(leaf_name, label=str(leaf))

		if isinstance(child, ProcessTree):
			draw_node(child, dot)
			child_name = ''.join([str(child.parent), str(child.id)])
		else:
			draw_leaf(parent, child, dot)
			child_name = ''.join([parent, child])

		# draw edge
		dot.edge(parent, child_name)
		return dot

	# general settings
	dot = Digraph(name, format=format)
	dot.attr(rankdir="TB", ranksep="equally")

	# draw root
	dot.attr('node', shape='circle', penwidth="1", fontsize="10",
			 fontname="Helvetica", height="0.3", fixedsize="true")

	node_name = ''.join(['root', str(tree.id)])
	dot.node(node_name, label=gen_label(tree.cut))

	for child in tree.children:
		dot = draw_children(node_name, child, dot)

	if render:
		render_dot(dot, name)

	return dot


def render_dot(dot: Digraph, name: str, cleanup=True, view=True):
	"""
	Render the given DOT object to file. After rendering, the image can be
	shown and artifacts can be removed.

	:param dot: DOT object
	:param name: file name
	:param cleanup: remove dot artifacts after rendering
	:param view: open file after rendering
	"""

	dot.render(filename=name, view=view, cleanup=cleanup)



"""
This package provides a class to build simple graph data structures,
like directly follows graph.
"""
from copy import copy, deepcopy
from collections import defaultdict
from pmlab_lite.pn import PetriNet


class Graph():

    def __init__(self):
        self.vertexes = defaultdict(list)
        self.start_nodes = []
        self.end_nodes = []
        self.collapsed_comp = {}

    def add_edge(self, o, t):
        if t not in self.vertexes[o]:
            self.vertexes[o].append(t)

        if t not in self.vertexes.keys():
            self.vertexes[t] = []


    def is_reachable(self, o, t):
        # set all nodes to unvisited
        nodes = list(self.vertexes.keys())
        visited = [False] * len(nodes)

        queue = []

        # vistit origin
        queue.append(o)
        visited[nodes.index(o)] = True

        while queue:
            n = queue.pop(0)
            if n == t:
                return True

            # loop all succiding vertexes
            for i in self.vertexes[n]:
                if not visited[nodes.index(i)]:
                    queue.append(i)
                    visited[nodes.index(i)] = True

        return False

    def from_log(self, log: list):
        for trace in log:
            if len(trace) >= 1:
                self.start_nodes.append(trace[0])
                self.end_nodes.append(trace[-1])
            self.from_seq(trace)

        return self

    def from_seq(self, seq: list):
        if len(seq) > 1:
            for o, t in zip(seq, seq[1:]):
                self.add_edge(o, t)
        elif len(seq) == 1:
            if seq[0] not in self.vertexes.keys():
                self.vertexes[seq[0]] = []

    def from_petrinet(self, pn: PetriNet):
        """
        Derive directly follows graph from Petri net.

        :param pn: PetriNet
        """
        l1 = list(filter(lambda x: not isinstance(x[0], int), pn.edges))
        for l in l1:
            for e in pn.edges:
                if l[1] == e[0]:
                    self.add_edge(l[0], e[1])

    def get_cc(self):

        def dfs(temp, node, visited, nodes):

            visited[nodes.index(node)] = True
            temp.append(node)

            for n in u_graph[node]:
                if visited[nodes.index(n)] == False:
                    temp = dfs(temp, n, visited, nodes)
            return temp

        cc = []

        # copy for undirected graph
        u_graph = deepcopy(self.vertexes)
        nodes = list(u_graph)
        visited = [False] * len(nodes)

        for k, v in u_graph.items():
            for n in v:
                if k not in u_graph[n]:
                    u_graph[n].append(k)

        for n in u_graph.keys():
            if visited[nodes.index(n)] == False:
                temp = []
                cc.append(dfs(temp, n, visited, nodes))

        return cc

    def get_scc(self):

        def build_stack(graph, node, nodes, visited, stack):
            if not visited[nodes.index(node)]:
                visited[nodes.index(node)] = True
                for succ in graph.vertexes[node]:
                    build_stack(graph, succ, nodes, visited, stack)

                stack.append(node)

        def transpose():
            g_t = Graph()
            for node, succ in self.vertexes.items():
                for s in succ:
                    g_t.add_edge(s, node)

            return g_t

        def visit(graph, node, nodes, visited, comp):
            if not visited[nodes.index(node)]:
                visited[nodes.index(node)] = True
                comp.append(node)
                for succ in graph.vertexes[node]:
                    visit(graph, succ, nodes, visited, comp)

        nodes = list(self.vertexes.keys())

        # 1. step
        stack = []
        visited = [False] * len(nodes)

        for n in self.vertexes:
            build_stack(self, n, nodes, visited, stack)

        # 2. step: invert graph
        g_t = transpose()

        # 3. step: find components
        visited = [False] * len(nodes)
        scc = []
        temp = []

        while stack:
            n = stack.pop()
            if not visited[nodes.index(n)]:
                visit(g_t, n, nodes, visited, temp)
                scc.append(temp)
                # reset comp
                temp = []

        return scc

    def invert_old(self):
        g_i = Graph()
        g_i.start_nodes = self.start_nodes
        g_i.end_nodes = self.end_nodes
        activities = self.vertexes.keys()
        for o in activities:
            for t in activities:
                if o != t:
                    if o not in self.vertexes[t] or t not in self.vertexes[o]:
                        g_i.add_edge(o, t)
                        g_i.add_edge(t, o)
                    else:
                        if o not in list(g_i.vertexes.keys()):
                            g_i.vertexes[o]

        return g_i

    def invert(self):
        g_i = Graph()
        g_i.start_nodes = self.start_nodes
        g_i.end_nodes = self.end_nodes
        activities = self.vertexes.keys()
        for o in activities:
            for t in activities:
                if o != t:
                    if o not in self.vertexes[t] and t in self.vertexes[o]:
                        g_i.add_edge(o, t)
                        g_i.add_edge(t, o)
                    else:
                        if o not in list(g_i.vertexes.keys()):
                            g_i.vertexes[o]

        return g_i

    # def get_start_nodes(self):
    #    nodes = list(self.vertexes.keys())
    #    succ = [act for sub in list(self.vertexes.values()) for act in sub]
    #    return list(set(nodes) - set(succ))

    # def get_end_nodes(self):
    #    return list(filter(lambda k: len(self.vertexes[k]) == 0,
    #                       self.vertexes.keys()))

    def remove_start(self):
        g_reduced = copy(self)
        new_ver = defaultdict(list)

        for node, succ in self.vertexes.items():
            if node not in self.start_nodes:
                if node not in new_ver.keys():
                    new_ver[node] = []
                for s in succ:
                    if s not in self.start_nodes:
                        new_ver[node].append(s)

        g_reduced.vertexes = new_ver

        return g_reduced

    def remove_end(self):
        g_reduced = copy(self)

        new_ver = defaultdict(list)

        for node, succ in self.vertexes.items():
            if node not in self.end_nodes:
                if node not in new_ver.keys():
                    new_ver[node] = []
                for s in succ:
                    if s not in self.end_nodes:
                        new_ver[node].append(s)

        g_reduced.vertexes = new_ver

        return g_reduced



from unittest import TestCase

from pmlab_lite.discovery.inductive_miner import InductiveMiner
from pmlab_lite.discovery.process_tree import ProcessTree
from pmlab_lite.helper.graph import Graph
from pmlab_lite.helper.viz import dot
from pmlab_lite.log import EventLog


class TestProcessTree(TestCase):

	def test_find_seq(self):
		log = [['a', 'x', 'b', 'c'],
			   ['a', 'x', 'c', 'b'],
			   ['a', 'x', 'd', 'e'],
			   ['a', 'x', 'd', 'e', 'f', 'd', 'e']]

		tree = ProcessTree(InductiveMiner(EventLog()), log, discover=False)

		split = tree.find_seq()
		self.assertListEqual(split, [['a', 'x']])
		self.fail()

	def test_find__max_seq(self):
		log = [['a', 'x', 'b', 'c'],
			   ['a', 'x', 'c', 'b'],
			   ['a', 'x', 'd', 'e'],
			   ['a', 'x', 'd', 'e', 'f', 'd', 'e']]

		tree = ProcessTree(InductiveMiner(EventLog()), log, discover=False)

		split = tree.find_seq()
		list(map(lambda sub: sub.sort(), split))
		for ele in split:
			for sub in ele:
				if isinstance(sub, list):
					sub.sort()
		# split.sort()
		self.assertListEqual(split, [['a'], ['x'], [['b', 'c'], ['d', 'e',
																 'f']]])


	def test_find_para(self):
		log = [['a', 'b'],
			   ['a', 'b'],
			   ['b', 'a']]

		# log = [['supervisor signature',  # i
		# 		'sign application'],  # j
		# 		['sign application',  # j
		# 		'supervisor signature']]  # i

		tree = ProcessTree(InductiveMiner(EventLog()), log, discover=False)

		split = tree.find_para()
		split.sort()
		print(split)
		self.assertListEqual(split, [['a'], ['b']])


	def test_find_excl(self):
		log = [['b', 'c'],
			   ['c', 'b'],
			   ['d', 'e'],
			   ['d' , 'e', 'f', 'd', 'e']]

		tree = ProcessTree(InductiveMiner(EventLog()), log, discover=False)

		split = tree.find_excl()
		list(map(lambda  sub: sub.sort(), split))
		split.sort()
		self.assertListEqual(split, [['b', 'c'], ['d', 'e', 'f']])

	def test_find_loop(self):
		log = [['d', 'e'],
			   ['d', 'e', 'f', 'g', 'd', 'e']]

		tree = ProcessTree(InductiveMiner(EventLog()), log, discover=False)

		split = tree.find_loop()
		list(map(lambda sub: sub.sort(), split))
		print(split)
		self.assertListEqual(split, [['d', 'e'], ['f', 'g']])


	def test_find_cuts(self):
		log4 = [['a', 'b', 'g', 'h', 'i', 'j', 'k'],
				['a', 'b', 'g', 'h', 'j', 'i', 'k'],
				['a', 'b', 'g', 'h', 'i', 'k'],
				['a', 'b', 'g', 'h', 'j', 'k'],
				['a', 'b', 'c', 'd', 'e', 'h', 'i', 'k'],
				['a', 'b', 'c', 'd', 'e', 'f', 'c', 'd', 'e', 'h', 'i', 'k']]

		log2 = [['a', 'b', 'c'],
				['a', 'c', 'b'],
				['a', 'd', 'e'],
				['a', 'd', 'e', 'f', 'd', 'e']]

		miner = InductiveMiner(EventLog())
		tree = ProcessTree(miner, log2)
		print()
		tree.print_tree()

		dot.draw_process_tree(tree)

	def test_scc(self):
		log = [['g'],
			   ['c', 'd', 'e'],
			   ['c', 'd', 'e', 'f', 'c', 'd', 'e']]

		g = Graph()

		g = g.from_log(log)

		print(g.start_nodes)
		print(g.vertexes)



import numpy
import heapq
import copy
import platform
import sys
import os
import json
import tqdm
import graphviz as gr
import pmlab_lite.alignments.Variables as v
import pmlab_lite.pn.pn as pn
import pmlab_lite.helper.io.pnml as pnml
import pmlab_lite.helper.io.xes as xes



class Alignment:
    def __init__(self):
        self.alignment_move=[]
        self.move_in_log=[]
        self.move_in_model=[]
        self.fitness=[]




class Astar (Alignment):

    def __init__(self):
        Alignment.__init__(self)
        self.solutions = []


    # A-star algorithm
    def Astar_Exe(self, model_path, log_path, destination_path = os.getcwd(), final_marking=['end'], no_of_solutions=1):
        #:param model_path: Address to the PNML file
        #:param log_path:  Address to the log file
        #:param destination_path: Destination address to write some files
        #:param final_marking: Final marking of the Petri net, like 'p8'
        #:return:


        # Reading the model
        total_trace, net, index_place_end, index_place_start, Incident_matrix_net = self.__Reading_IO(model_path, log_path)


        #All the solutions, i.e. alignments
        v.solutions = dict()
        for key in tqdm.tqdm(total_trace):
            trace = total_trace[key]

            # Creating Synchronous products
            elements_tot, places, index_place_start_log, index_place_end_log, Incident_matrix = self.__Synchronous_Product(net, trace)

            # Creating initial marking vector
            init_mark_vector = list(numpy.repeat(0, len(places)))
            init_mark_vector[index_place_start] = 1
            init_mark_vector[index_place_start_log] = 1

            # Creating Final marking vector
            final_mark_vector = list(numpy.repeat(0, len(places)))
            final_mark_vector[index_place_end] = 1
            final_mark_vector[index_place_end_log] = 1
            final_marking = list(net.places.values())[index_place_end]

            # Storing the specification of SP net
            v.Incident_matrix = Incident_matrix
            v.init_mark_vector = init_mark_vector
            v.final_mark_vector = final_mark_vector



            # ---Initializing variables
            v.closed_list = closed_list = []
            #v.solutions = solutions = []
            v.open_list = open_list = []
            self.solutions=[]
            self.fitness=[]
            self.move_in_log=[]
            self.move_in_model=[]
            self.alignment_move=[]

            # Creating initial node in the search space
            current_node = Node()
            current_node.marking_vector = numpy.array(init_mark_vector[:])
            current_node.observed_trace_remain = trace
            open_list.append([current_node.total_cost, current_node])
            current_node.Heuristic_to_Final()


            #Iterating until open_list has an element
            while (len(open_list) > 0):
                # Sorting openlist
                heapq._heapify_max(open_list)

                current_node = heapq.heappop(v.open_list)
                closed_list.append(current_node)
                current_node = current_node[1]

                # Investigating the current node
                current_node.Investigate(Incident_matrix, elements_tot, places, final_marking)


                # Check weather the node if a solution or no:
                if (numpy.array_equal(current_node.marking_vector, final_mark_vector)):
                    #solutions.append(current_node)
                    self.solutions.append(current_node)
                    self.__Fitness()
                    self.__Move_alignment()
                    self.__Move_in_log()
                    self.__Move_in_model()
                    v.solutions[key]=[{'alignment': self.alignment_move, 'fitness': self.fitness}]
                    #print("Solution was found:", key)
                    if (len(self.solutions) >= no_of_solutions):
                        break


        #Write the output as a Json file
        with open('Results.json', 'w') as outfile:
            json.dump(v.solutions, outfile)
            print("Done!")

            # return trans, places, Incident_matrix , initial_place_marking, init_mark_vector, candidate_nodes, total_trace, distance_matrix, vertices

            #return solutions

    #########################################################################################
    # Reading input eventlog and model file (private method using "__")
    def __Reading_IO(self,model, log):
        '''
        :param model: direction to a PNML file
        :param log:  direction to a XES file
        :return: traces, net, incidence_matrix
        '''

        # Reading log from XES file
        def Read(log):
            if(log.endswith(".xes")):
                traces = dict()
                obj = xes.import_from_xes(log)
                for case in obj.get_traces():
                    for c in case:
                        if (c.get_case_id() not in traces):
                            traces[c.get_case_id()] = [c.get_activity_name()]
                        else:
                            traces[c.get_case_id()].append(c.get_activity_name())

                            # print(c.get_activity_name(), c.get_case_id())
            # Reading log from a text file!
            elif (log.endswith(".txt")):
                f = open(log)
                traces=dict()
                count=0
                for line in f:
                    #trace = f.read()
                    trace = line.strip('\n').split(' ')
                    # for i in range(len(trace)):
                    # trace[i] = "T_" + trace[i]
                    traces[count]=trace
                    count+=1

            return traces

        # Creating Incidence matrix
        def Incidence_Matrix(net):
            # Creating an empty matrix
            incidnet_matrix = numpy.zeros((len(net.places.keys()), len(net.transitions.keys())), dtype=int)

            for keyT in net.transitions.keys():
                for keyP in net.places.keys():
                    if (net.transitions[keyT][0], net.places[keyP]) in net.edges:
                        col_index = list(net.transitions.values()).index(net.transitions[keyT])
                        row_index = list(net.places.values()).index(net.places[keyP])
                        incidnet_matrix[row_index][col_index] = 1
                    elif (net.places[keyP], net.transitions[keyT][0]) in net.edges:
                        col_index = list(net.transitions.values()).index(net.transitions[keyT])
                        row_index = list(net.places.values()).index(net.places[keyP])
                        incidnet_matrix[row_index][col_index] = -1

            return incidnet_matrix

        traces = Read(log)

        net = pn.PetriNet()
        pnml.load(net, model)
        incidence_matrix = Incidence_Matrix(net)

        # Finding the final marking
        # Determining which row has 0 or positive elements (that is the final marking place)
        index_place_end = numpy.where((incidence_matrix >= 0).all(axis=1))
        # index_place_end=numpy.where(numpy.sum(incidence_matrix, axis=1) == 1)
        if (len(index_place_end) > 1):
            raise ValueError('place identifier has to be unique')

        # Finding the initial marking markin
        # Determining which row has 0 or negative elements (that is the initial marking place)
        index_place_start = numpy.where((incidence_matrix <= 0).all(axis=1))
        # index_place_start = numpy.where(numpy.sum(incidence_matrix, axis=1) <= -1)
        if (len(index_place_end) > 1):
            raise ValueError('place identifier has to be unique')


        return traces, net, index_place_end[0][0], index_place_start[0][0], incidence_matrix

    ######################################################################################
    # Crating synchronous product of a model an log (private using "__")
    class __elements:
        def __init__(self, id, type, index):
            self.place_pre = []
            self.place_post = []
            self.trans_id = id
            self.trans_type = type  # This can be synchronous, model or log
            self.index = index  # Index of elements in the trace or model

    def __Synchronous_Product(self, net, trace):
        '''

        :param net: is an object from Petri net class
        :param trace: an input trace, for which sych product will be created
        :return: sync product
        '''
        # net.edges is like [(5, -3),(-2, 5),(6, -4),..]
        # net.transition is like {'t4': [-1],'t5': [-2],'t6': [-3],....}
        elements_tot = []
        t_model = list(net.transitions.keys())
        for i in range(len(t_model)):
            e = self.__elements(t_model[i], "model", i)
            for pair in net.edges:
                if (net.transitions[t_model[i]][0] == pair[0]):
                    e.place_post.append(pair[1])
                elif (net.transitions[t_model[i]][0] == pair[1]):
                    e.place_pre.append(pair[0])
            elements_tot.append(e)

        # Creating trace net
        counter = 0
        for i in range(len(trace)):
            e = self.__elements(trace[i], "log", i)
            e.place_pre.append("PL" + str(counter))
            counter += 1
            e.place_post.append("PL" + str(counter))
            elements_tot.append(e)

        final_log_place = "PL" + str(counter)
        initial_log_place = "PL" + str(0)

        # Creating synchronous transitions
        for i in range(len(trace)):
            e = self.__elements(trace[i], "Sync", i)
            for node in elements_tot:
                if (trace[i] == node.trans_id):
                    if ((node.trans_type == 'log' and node.index != i) or (node.trans_type == 'Sync')):
                        pass
                    else:
                        e.place_pre = e.place_pre + node.place_pre
                        e.place_post = e.place_post + node.place_post
            elements_tot.append(e)

        # Creating incidence matrix of synchronous product
        places = list(net.places.values())
        # places=set()
        for e in elements_tot:
            for p in e.place_pre:
                if (p not in places):
                    places.append(p)
            for p in e.place_post:
                if (p not in places):
                    places.append(p)
        # places = list(places)


        index_place_start_log = places.index(initial_log_place)
        index_place_end_log = places.index(final_log_place)



        # Initializing and filling in the matrix
        incidence_matrix = numpy.zeros((len(places), len(elements_tot),), dtype=int)
        for i in range(len(elements_tot)):
            for p in elements_tot[i].place_pre:
                row_ind = places.index(p)
                incidence_matrix[row_ind][i] = -1
            for p in elements_tot[i].place_post:
                row_ind = places.index(p)
                incidence_matrix[row_ind][i] = 1

        # Plotting synchronous product model
        #Drawing_Model(elements_tot, places, incidence_matrix)
        #Storing variables as a Private ones (using "__")
        self.__elements_tot=elements_tot
        self.__places=places
        self.__incidence_matrix=incidence_matrix

        return elements_tot, places, index_place_start_log, index_place_end_log, incidence_matrix
#####################################################################################################
    def __Fitness(self):
        for sol in self.solutions:
            u = sol.alignment_Up_to
            self.fitness.append(round(len([e for e in u if (e[0] != '-' and e[1] != '-')]) / len(u), 3))
    def __Move_alignment(self):
        for sol in self.solutions:
            self.alignment_move.append(sol.alignment_Up_to)
    def __Move_in_model(self):
        for sol in self.solutions:
            u=sol.alignment_Up_to
            self.move_in_model.append([e[1] for e in u if e[1] != '-'])
    def __Move_in_log(self):
        for sol in self.solutions:
            u=sol.alignment_Up_to
            self.move_in_log .append([e[0] for e in u if e[0] != '-'])
    ################################################################################################
    # This method plots the synchronous products
    def Drawing_Model(self):
        places=self.__places
        incident_matrix=self.__incidence_matrix
        elements_tot=self.__elements_tot
        # path=V.destination_path

        # Now attempting to create a graph

        dot = gr.Digraph()

        for i in range(len(places)):
            dot.node(str(places[i]), shape="circle")

        # Here during the initializing the transition nodes of the graph, if we encountered the transition which is appeared in the net_moves,
        # we highlight it and also assign it a moves number.
        # Similarly transitions which are asynch moves or Skipped, will receive another colors
        for e in elements_tot:
            index = str(e.index)
            if e.trans_type == 'model':
                dot.node(str(e.trans_id + "-" + e.trans_type + index), shape="rect", style='filled', color="red")

            elif e.trans_type == 'log':
                dot.node(str(e.trans_id + "-" + e.trans_type + index), shape="rect", style='filled', color="green")
            elif e.trans_type == 'Sync':
                dot.node(str(e.trans_id + "-" + e.trans_type + index), shape="rect", style='filled', color="blue")


                # else:
                # dot.node(transitions[i], shape="rect")

        for i in range(len(incident_matrix)):
            for j in range(len(incident_matrix[0])):
                index = str(elements_tot[j].index)
                if (incident_matrix[i][j] == 1):

                    dot.edge(str(elements_tot[j].trans_id + "-" + elements_tot[j].trans_type + index), str(places[i]))


                elif (incident_matrix[i][j] == -1):

                    dot.edge(str(places[i]), str(elements_tot[j].trans_id + "-" + elements_tot[j].trans_type + index))

                    # Graphviz must be installed in order to save output in pdf format
        if platform.system() == "Windows":
            os.environ["PATH"] += os.pathsep + 'C:/Program Files (x86)/Graphviz2.38/bin/'
        elif platform.system() == "Linux":
            if ('graphviz' not in sys.modules):
                os.system('sudo apt-get install graphviz')

        print ("the net move is figured")
        f = open(os.getcwd() + "/Graph_net_moves_color.dot", "w")
        dot.render("plot.png")
        f.write(dot.source)

########################################################################################################
#######################################################################################################
#This is a class that creates the state space of SPN (called from Astar)
class Node():
    def __init__(self):
        #Astar.__init__(self)

        # Shows a marking node
        self.marking_vector = []
        self.active_transition = []
        self.parent_node = ''
        self.observed_trace_remain = []
        self.alignment_Up_to = []  # It is of the form ((t1,t1),(t2,-),(-,t3))
        self.cost_from_init_marking = sum([0 if (x[0] and x[1]) != '-' else 1 for x in
                                           self.alignment_Up_to])  # Computing steps weight in an alignment
        self.cost_to_final_marking = 1000
        self.total_cost = self.cost_to_final_marking + self.cost_from_init_marking

    # Methods for comparison two objects (Only python 3.7)
    def __lt__(self, other):
        return self.total_cost < other.total_cost

    def Find_active_transition(self, incidence_matrix, elements_tot):
        # Finding active transitions
        # Looping over transitions of the model to see which one is active given the marking of that node
        for i in range(0, incidence_matrix.shape[1]):
            if (numpy.all((incidence_matrix[:, i] + self.marking_vector) >= 0)):
                if (i not in self.active_transition):
                    self.active_transition.append(i)


                    ##################################################################

    # This function calls other functions to investigate the current node
    def Investigate(self, incidence_matrix, elements_tot, places, final_marking):
        '''

        :param incidence_matrix: Incidence matrix of the Petri net (numpy index)
        :param trans: A list of transitions of the model
        :param places: A list of places of the model
        :param final_marking: Final marking of the model, like "p8'
        :param distance_matrix: Distance matrix computed from Floyed algorithm
        :param vertices: Names of vertices of the distance matrix
        :return:
        '''

        # Finding active transitions
        self.Find_active_transition(incidence_matrix, elements_tot)


        # Heuristic evaluation of active transitions
        for i in self.active_transition:
            # Computing the distance of an active transition to the final marking (a place in case of WF net)
            self.Heuristic_to_Final()

            # ------------------------Movements-------------------
            # --Synchronous
            if (elements_tot[i].trans_type == "Sync"):

                # Creating a new node
                node_child = Node()


                # Updating the current marking for that node by firing trans[i]
                node_child.marking_vector = self.marking_vector + incidence_matrix[:, i]
                node_child.parent_node = self

                # Updating the remaining observed trace
                node_child.observed_trace_remain = self.observed_trace_remain[1:]
                node_child.alignment_Up_to = self.alignment_Up_to + [
                    (self.observed_trace_remain[0], elements_tot[i].trans_id)]
                # node_child.cost_to_final_marking = d
                node_child.Heuristic_to_Final()
                node_child.cost_from_init_marking = 1 * sum(
                    [1 if ((x[0] != '-') and (x[1] != '-'))  else 0 for x in
                     node_child.alignment_Up_to]) / float(
                    len(node_child.alignment_Up_to))

                # Updating active transitions for the child node
                node_child.total_cost = node_child.cost_to_final_marking + node_child.cost_from_init_marking

                self.Add_node(v.open_list, v.closed_list, node_child, ",synch,")



            # Move in model
            elif (elements_tot[i].trans_type == "model"):
                if (numpy.all(self.marking_vector + incidence_matrix[:, i] >= 0)):

                    # Creating a new node
                    node_child = Node()
                    # Updating the current marking for that node by firing trans[i]
                    node_child.marking_vector = self.marking_vector + incidence_matrix[:, i]
                    #print("Marking of child in move in model:", self.marking_vector + incidence_matrix[:, i])
                    node_child.parent_node = self

                    # Updating the remaining observed trace
                    node_child.observed_trace_remain = self.observed_trace_remain[:]
                    node_child.alignment_Up_to = self.alignment_Up_to + [('-', elements_tot[i].trans_id)]
                    # node_child.cost_to_final_marking = d
                    node_child.Heuristic_to_Final()
                    node_child.cost_from_init_marking = 1 * sum(
                        [1 if ((x[0] != '-') and (x[1] != '-'))  else 0 for x in
                         node_child.alignment_Up_to]) / float(
                        len(node_child.alignment_Up_to))

                    # Updating active transitions for the child node
                    node_child.total_cost = node_child.cost_to_final_marking + node_child.cost_from_init_marking


                    # Checking whether it is in the closed list

                    self.Add_node(v.open_list, v.closed_list, node_child, ",move in model,")



            # Move in log
            elif (elements_tot[i].trans_type == "log"):


                # Creating a new node
                node_child = Node()
                # Updating the current marking for that node
                node_child.marking_vector = self.marking_vector + incidence_matrix[:, i]
                #print("Marking of child in move in model:", self.marking_vector + incidence_matrix[:, i])
                node_child.parent_node = self

                # Updating the remaining observed trace
                node_child.observed_trace_remain = self.observed_trace_remain[1:]
                node_child.alignment_Up_to = self.alignment_Up_to + [(self.observed_trace_remain[0], '-')]
                # node_child.cost_to_final_marking = self.cost_to_final_marking
                node_child.Heuristic_to_Final()
                node_child.cost_from_init_marking = 1 * sum(
                    [1 if ((x[0] != '-') and (x[1] != '-'))  else 0 for x in
                     node_child.alignment_Up_to]) / float(
                    len(node_child.alignment_Up_to))

                # Updating active transitions for the child node
                node_child.total_cost = node_child.cost_to_final_marking + node_child.cost_from_init_marking


                # Checking whether it is in the closed list
                self.Add_node(v.open_list, v.closed_list, node_child, ',move in log,')




                    #############################################################################################################

    # Heuristic that estimates from the current node to the final marking how many transitions must be fired!
    def Heuristic_to_Final(self):

        b = numpy.array(v.final_mark_vector) - numpy.array(self.marking_vector)
        x = numpy.linalg.lstsq(v.Incident_matrix, b, rcond=None)[0]
        x[x > 0] = 1
        x[x <= 0] = 0

        self.cost_to_final_marking = 1 / numpy.sum(x)
    #-----------------
        # Deciding whether to add a now node to the open list
    def Add_node(self, open_list, closed_list, node_child, id):
            # Checking whether it is in the closed list

            ind = [k for k in range(len(closed_list)) if
                   numpy.array_equal(node_child.marking_vector, closed_list[k][1].marking_vector)]

            if (len(ind) > 0):
                pass


            # checking whether it is in the open list (update if we found it)
            else:

                # ind is a index list like [12,34,10]
                ind = [k for k in range(len(open_list)) if
                       numpy.array_equal(node_child.marking_vector, open_list[k][1].marking_vector)]

                if (ind):
                    for k in ind:
                        if (open_list[k][1].cost_from_init_marking < node_child.cost_from_init_marking):
                            open_list[k] = [node_child.total_cost, node_child]
                        else:
                            continue
                else:
                    open_list.append([node_child.total_cost, node_child])

                    ###########################################################



